### Persephone - a socio-economic-ecological model of European agricultural landscapes.
###
### This file manages the landscape maps that underlie the model.
###

"The land cover classes encoded in the Mundialis Sentinel data."
@enum LandCover nodata forest grass water builtup soil agriculture
## Do not change the order of this enum, or initlandscape() will break!

"The types of landscape event that can be simulated"
@enum EventType tillage sowing fertiliser pesticide harvest

"""
    Pixel

A pixel is a simple data structure to combine land use and ownership information
in a single object. The model landscape consists of a matrix of pixels.
(Note: further landscape information may be added here in future.)
"""
mutable struct Pixel
    landcover::LandCover
    fieldid::Union{Missing,Int64}
    events::Vector{EventType}
end

"""
    FarmEvent

A data structure to define a landscape event, giving its type,
spatial extent, and duration.
"""
mutable struct FarmEvent
    type::EventType
    pixels::Vector{Tuple{Int64,Int64}}
    duration::Int64
end

"""
    initlandscape()

Initialise the model landscape based on the map files specified in the
configuration. Returns a matrix of pixels.
"""
function initlandscape()
    @debug "Initialising landscape"
    landcover = GeoArrays.read(param("core.landcovermap"))
    farmfields = GeoArrays.read(param("core.farmfieldsmap"))
    (size(landcover) != size(farmfields)) && Base.error("Input map sizes don't match.")
    width, height = size(landcover)[1:2]
    landscape = Matrix{Pixel}(undef, width, height)
    for x in 1:width
        for y in 1:height
            # convert the numeric landcover value to LandCover, then create the Pixel object
            lc = landcover[x,y][1]
            (ismissing(lc)) && (lc = 0)
            lcv = LandCover(Int(lc/10))
            ff = farmfields[x,y][1]
            !(ismissing(ff)) && (ff = Int64(ff))
            landscape[x,y] = Pixel(lcv, ff, Vector{Symbol}())
        end
    end
    return landscape
end

"""
    updateevents!(model)

Cycle through the list of events, removing those that have expired.
"""
function updateevents!(model::AgentBasedModel)
    expiredevents = []
    for e in 1:length(model.events)
        event = model.events[e]
        event.duration -= 1
        if event.duration <= 0
            push!(expiredevents, e)
            for p in event.pixels
                i = findnext(x -> x == event.type, model.landscape[p...].events, 1)
                deleteat!(model.landscape[p...].events, i)
            end
        end
    end
    deleteat!(model.events, expiredevents)
end

"""
    createevent(model, pos, name, duration=1)

Add a farm event to the specified pixels for a given duration.
"""
function createevent!(model::AgentBasedModel, pixels::Vector{Tuple{Int64,Int64}},
                     name::EventType, duration::Int64=1)
    push!(model.events, FarmEvent(name, pixels, duration))
    for p in pixels
        push!(model.landscape[p...].events, name)
    end
end

"""
    landcover(model, position)

Return the land cover class at this position (utility wrapper).
"""
function landcover(pos::Tuple{Int64,Int64}, model::AgentBasedModel)
    model.landscape[pos...].landcover
end

"""
    farmplot(model, position)

Return the farm plot at this position, or nothing if there is none (utility wrapper).
"""
function farmplot(pos::Tuple{Int64,Int64}, model::AgentBasedModel)
    ismissing(model.landscape[pos...].fieldid) ? nothing :
              model[model.landscape[pos...].fieldid]
end

"""
    distanceto(model, pos, habitattype)

Calculate the distance from the given location to the closest habitat of the specified type.
Caution: can be computationally expensive!
"""
function distanceto(pos::Tuple{Int64,Int64}, model::AgentBasedModel, habitattype::LandCover)
    #XXX can I make this check for both land cover type and crop type?
    # (Or even take a full habitat descriptor?)
    #TODO
end

"""
    distancetoedge(model, pos)

Calculate the distance from the given location to the closest neighbouring habitat.
Caution: can be computationally expensive!
"""
function distancetoedge(pos::Tuple{Int64,Int64}, model::AgentBasedModel)
    lc = landcover(model, pos)
    dist = 1
    #TODO
end

